#################################################
# Starsiege Terrain Exporter v1.0 8-03-04		#
# Copyright  2004 Super Powio			#
# Powio@hotmail.com					#
# www.SunAndShadows.com/MIB				#
#								#
# Special Thanks:						#
# ---------------						#
# Sentinal [MIB] - Beta Testing			#
# Raven [MIB] - Pointing out that I didn't need	#
# 	a new exterior program				#
# 								#
# Michael Johnston - Original Tribes Script and	#
#	program (the one that's used here)		#
#################################################

// This script and the included program are used, in conjunction, to make the
// 16-bit grayscale height maps that Unreal and some other games use to define
// terrain.
//
// The script will scan the terrain and dump raw heights into a text
// dump called 'TerrainExporter.log'.
//
// ConvTED.exe will read the text dump and convert it into a 16-bit raw image
// file, which can then be read by Adobe Photoshop or a game engine (such as
// Unreal, Tribes 2, and Tribes: Vengeance).
//
// REQUIREMENTS: (Mostly taken from Michael Johnston's read-me file. Whups. ^_^ )
// 1.) ConvTED.exe, written by Michael "KineticPoet" Johnston. This is necessary for converting the text dump to a raw binary file.
// 2.) A program that can interpret raw binary bitmaps as images.
//	a) Adobe Photoshop.  Open your .raw file.  Set the import options to 16-bits, 1 channel, IBM-PC.
//	b) Daylon Leveller.  Import a raw binary file. Set options to 512x512, unsigned 16-bit integers, Intel, 0 byte header. I haven't
//		tested this to see if it supports the extra set of pixels from using the 'edges' argument in the script.
//		Leveller is available at: http://www.daylongraphics.com/products/leveller/
//	c) G16ed. This one is a height map editor for Unreal. Import a RAW image, 16 bits per pixel, Intel, 0 byte header. This one won't
//		support images generated using the 'edges' argument from the script. Available at: http://homepage.ntlworld.com/martingbell/
//
// INSTRUCTIONS:
// 1.  Copy this script to Starsiege's scripts folder, or its main directory.
// 2.  Delete TerrainExporter.log if it exists. You should always do this before exporting a new terrain or it will become messy.
// 3.  Start up Starsiege. If your console isn't already enabled, click the small white spot on the gold border of the main menu. It should
//	 be directly below the period of the version number in the lower left corner of the screen. Load the level whose terrain you
//	 want to export.
// 4.  Open up the console with the '~' key. Type in 'exec("exportTerrain");'. Without the outer quotes, of course.
// 5.  Type in 'exportTerrain();' for a list of arguments, which are explained below. For now, just type in 'exportTerrain(false);' if you
//	 don't understand.
// 6.  Your game should seem to freeze up. Don't worry, this is normal. Depending on the resolution you chose for the command, this could
//	 take a long time (up to 1-4 HOURS for resolutions 2 and 3!). Since the resolution is defaulted to 0, this should only take three
//	 minutes or so.
// 7.  Do NOT refresh your Starsiege folder. Do NOT open or even CLICK the newly generated TerrainExporter.log file while the script is
//	 still processing. You'll corrupt the file.
// 8.  When the script is done, you'll get an audio notification (unless Starsiege is running the background). Shut down the game, and look
//	 for TerrainExporter.log in your main Starsiege directory. Open it up using your favorite text editor.
// 9.  Remove the header and footer of the file (the lines that say BEGIN and END) so that you're left with just the numbers. Save that file
//	 in the same directory as ConvTED.exe.
// 10. Open up a DOS prompt, and navigate to where you extracted ConvTED.exe. Use the parameters:
//     'ConvTED.exe TerrainExporter.log myMap.raw'. This will convert the text dump into a binary file called myMap.raw.
// 11. If you saved your log as something else, enter its name in place of console.log.  Replace myMap.raw with whatever name you want.
// 12. Open the .raw file using one of the aforementioned programs above. You exported your terrain with the 'edges' argument, open is using
//	 Adobe Photoshop, making sure you add one extra pixel to the dimensions of your image.
// 13. Edit and/or save your image to the file format that the editor you're importing to can read. Probably .bmp or .jpg.
// 14. Import your image into the BSP editor of your choice. Voila.
//
// USAGE:
// exportTerrain( edges , [resolution], [section], [zParams] ):	Scans the terrain and dumps a 16-bit height list into TerrainExporter.log
//	edges			: Defines whether or not to scan the right and bottom edges of the area, for better accuracy. This option was added
//				  for the perfectionists, so the exporter doesn't lop off a small strip of land bordering the right and bottom edge
//				  of the scanning area. This argument, when set to TRUE, will add one extra pixel to the dimensions of the image.
//				  Note that when using this option, you will most likely have to edit the height map through a 3rd party program
//				  that supports 16-bit images, such as Adobe Photoshop (Jasc Paint Shop Pro 8 doesn't support 16-bit images).
//
//	resolution:	Changes the height map size in exponents of 2. Ranges from -4 - 3. Default is 0 (256 x 256).
//		-4 : 16 x 16
//		-3 : 32 x 32
//		-2 : 64 x 64
//		-1 : 128 x 128
//		 0 : 256 x 256
//		 1 : 512 x 512 (From here on, scanning takes a LONG time!)
//		 2 : 1024 x 1024
//		 3 : 2048 x 2048
//
//	section:	Defines which section of the terrain you want to export. 
//		default 	: Scans whatever's under the SAT map. Useful if the map has changed the SAT Map's position.
//		main		: Scans the main area of the terrain. This is the normal play area of the map.
//		secondary	: Scans the outer area of the terrain. This is the area outside the normal play area that repeats itself eight times around the map.
//		custom	: Defined by the variables $TerrainExporter::CustomXOffset and $TerrainExporter::CustomYOffset. Scans one SAT Map area starting from this point.
//
//	zParams:	Defines how the script should grab the minimum and maximum height of the terrain.
//		default	: Scans both the main area and the secondary area of the terrain for the min and max height.
//		area		: Scans the area being exported for the min and max height.
//		main		: Scans only the main area of the terrain for the min and max height.
//		secondary	: Scans only the secondary area of the terrain for the min and max height.
//		custom	: Uses the min and max heights defined by the variables $TerrainExporter::zMin and $TerrainExporter::zMax.
//
// $TerrainExporter::$CustomXOffset	: Defines the starting coordinates to start scanning the terrain at, when using the 'custom' option
// $TerrainExporter::$CustomYOffset	  for the 'section' argument.
// $TerrainExporter::$CustomZMin	: Defines the minimum and max heights of the terrain, necessary for creating a proper grayscale height
// $TerrainExporter::$CustomZMax	  map. For use with 'custom' option for the 'zParams' argument.
// $server::HudMapViewOffsetX		: These variables are built into the game. They change the center point of the SAT Map.
// $server::HudMapViewOffsetY
//
// TIPS:
// - When exporting the terrain, the game may seem to freeze. Don't panic, this is normal. It can take up to hours to export the terrain
//   at a higher resolution, but the default setting should take no longer than three minutes. By the way, don't touch the exported file
//   or folder while the script is still processing the height data.
// - When the 'edges' argument is set to true, it will increase the dimensions of the height map by one. Remember to add one extra pixel
//   to the dimensions of your image after using this option.
// - Sometimes the SAT Map will not always be in the center of the map, for some maps. When you're going to grab the main area of the
//   terrain, make sure you use the 'main' option for the 'section' argument.
// - Avoid exceeding -12288, or 12288 for either the X or Y coordinate during the scan. Doing so may produce an incorrect height map.
//
// - USEFUL TIP: If you have Adobe Photoshop, or any other imaging program that supports 16-bit images, you can drastically cut the time
//   for exporting a height map, while still retaining an accurate to the edge, and seamless height map. Here's how:
//
//   1. Export the terrain with the 'edges' argument set to true, and resolution at 0. Use whatever other parameters you like. This will
//	  create a text dump for a 257 x 257 pixel image.
//   2. Convert the text dump as you would normally. Open up the output file in Adobe Photoshop, setting the dimensions of the image to
//      257 x 257.
//   3. Resize your image to something like 1025 x 1025 using a bilinear or weighted-average filter. Your height map should appear very
//      smooth-like.
//   4. Crop (not resize!) the image to one less pixel than the dimensions you chose. This should take out the excess pixels on the right
//      and bottom border. Save your image.
//   5. You now have the most accurate representation you will ever have of the terrain you just exported, and fast, too.
//
//   You can also export the .raw file from G16er as an 8-bit grayscale so that you can use it with PSP, but note that you'll be losing
//   the recommended 16-bits that Unreal uses to differentiate between pixels of the same lightness.
//-----------------------------------------

function exportTerrain( %edges, %resolution, %section, %zParams )
{
	%resolution = floor(%resolution);	// Default 0 if not numeric

	// User does not input arguments correctly
	if( %edges != FALSE && %edges != TRUE )
	{
		echo( "exportTerrain( edges (true/false), [resolution (default:0)], [section (default/main/secondary/custom)], [z-params (default/area/main/secondary/custom)] )" );
		return;
	}

	// User does not input a valid resolution
	if( %resolution < -4 || %resolution > 3 )
	{
		echo( "exportTerrain: resolution must be between -4 and 3 (default:0)!" );
		return;
	}

	TerrainExporter::initRanges(%edges, %section, %resolution);
	TerrainExporter::initZRange(%edges, %zParams);

	// Check to see if the Z-Ranges are valid before exporting the terrain
	if($TerrainExporter::zMax <= $TerrainExporter::zMin)
	{
		echo("ERROR: zMax is less than or equal to zMin. Aborting...");
		return;
	}

	// Add a little more info to the log information string
	$TerrainExporter::LogInfo = strCat($TerrainExporter::LogInfo, "(Z-Min:"@ $TerrainExporter::zMin @" Z-Max:"@ $TerrainExporter::zMax @") " );

	// Okay, we're ready. Let's go.
	TerrainExporter::scanTerrain(%edges);
}

function TerrainExporter::initRanges( %edges, %section, %resolution )
{
	// Calculate the dimensions of the output height map based on resolution input
	%mapSize = 8192;
	%heightMapSize = 256 * raiseToPower(2,%resolution);

	// Calculate the distance between points the terrain will be scanned at
	$TerrainExporter::increment = %mapSize / %heightMapSize;

	// The terrain is divided into two distinct sections; set the area to export
	if(%section == main)
	{
		$TerrainExporter::xStart = -4096;
		$TerrainExporter::yStart = 4096;
	}
	else if(%section == secondary)
	{
		$TerrainExporter::xStart = -4096;
		$TerrainExporter::yStart = 12288;
	}
	else if(%section == custom)
	{
		$TerrainExporter::xStart = floor($TerrainExporet::CustomX);
		$TerrainExporter::yStart = floor($TerrainExporet::CustomY);
	}
	else	// Default: Satellite Map
	{
		%section = Satellite;

		$TerrainExporter::xStart = floor($server::HudMapViewOffsetX - 4096);
		$TerrainExporter::yStart = floor($server::HudMapViewOffsetY + 4096);
	}
		$TerrainExporter::xEnd = $TerrainExporter::xStart + %mapSize;
		$TerrainExporter::yEnd = $TerrainExporter::yStart - %mapSize;

	// If %edges is true, the exporter will add one to the dimensions of the height map
	if(%edges)
		%heightMapSize = strCat(%heightMapSize, "+1");

	// Set up the log information
	$TerrainExporter::LogInfo = $MissionName @" ("@ %section @") ["@ %heightMapSize @"x"@ %heightMapSize @"] ";
}

function TerrainExporter::initZRange(%edges, %zParams)
{
	// Reset the Z Ranges
	$TerrainExporter::zMin = "";
	$TerrainExporter::zMax = "";

	// If the user has entered Z-Parameters for the task, skip straight to exporting
	if(%zParams == custom)
	{
		$TerrainExporter::zMin = floor($TerrainExporter::CustomZMin);
		$TerrainExporter::zMax = floor($TerrainExporter::CustomZMax);
		return;
	}
	else if(%zParams == area)
	{
		%xScanStart = $TerrainExporter::xStart;
		%yScanStart = $TerrainExporter::yStart;

		%xScanEnd = $TerrainExporter::xEnd;
		%yScanEnd = $TerrainExporter::yEnd;
	}
	else if(%zParams == main)
	{
		%xScanStart = -4096;
		%yScanStart = 4096;

		%xScanEnd = 4096;
		%yScanEnd = -4096;
	}
	else if(%zParams == secondary)
	{
		%xScanStart = -4096;
		%yScanStart = 12288;

		%xScanEnd = -4096;
		%yScanEnd = 4096;
	}
	else		// Default: main and secondary sections
	{
		%xScanStart = -4096;
		%yScanStart = 12288;

		%xScanEnd = 4096;
		%yScanEnd = -4096;
	}

	// Now scan for the Min and Max heights
	for(%y = %yScanStart; %y >= %yScanEnd; %y -= $TerrainExporter::increment)
	{
		for(%x = %xScanStart; %x <= %xScanEnd; %x += $TerrainExporter::increment)
		{
			%y = round(%y);	// Starsiege sucks at handling decimals. This is after scanning buggy point 12288.

			if(%x == 12288)	// Point 12288 on the terrain is considered empty space, therefore we compensate
				%x -= 0.01;
			if(%y == 12288)
				%y -= 0.01;

			%height = getTerrainHeight(%x,%y);

			if(%height < $TerrainExporter::zMin || $TerrainExporter::zMin == "")
				$TerrainExporter::zMin = %height;
			if(%height > $TerrainExporter::zMax || $TerrainExporter::zMax == "")
				$TerrainExporter::zMax = %height;
		}
	}
}

function TerrainExporter::scanTerrain(%edges)
{
	fileWrite( "TerrainExporter.log", append, strCat( "***BEGIN TERRAIN HEIGHTS*** ", $TerrainExporter::LogInfo, getTime()) );

	// The Scanner
	for(%y = $TerrainExporter::yStart; %y >= $TerrainExporter::yEnd; %y -= $TerrainExporter::increment)
	{
		if(%y == $TerrainExporter::yEnd && !%edges)	// If edges isn't true, escape from the loop
		{
			break;
		}

		for(%x = $TerrainExporter::xStart; %x <= $TerrainExporter::xEnd; %x += $TerrainExporter::increment)
		{

			if(%x == $TerrainExporter::xEnd && !%edges)	// If edges isn't true, escape from the loop
			{
				break;
			}

			%y = round(%y);	// Starsiege sucks at handling decimals. This is after scanning buggy point 12288.

			if(%x == 12288)	// Point 12288 on the terrain is considered empty space, therefore we compensate
				%x -= 0.01;
			if(%y == 12288)
				%y -= 0.01;

			%height = getTerrainHeight(%x,%y);

			// The input ranges from 0 - 65535
			%input = round( ((%height - $TerrainExporter::zMin) / ($TerrainExporter::zMax - $TerrainExporter::zMin)) * 65535);

			if(%input > 65535)
				%input = 65535;
			if(%input < 0)
			{
				%input = 0;
			}

			// This becomes the output
			fileWrite( "TerrainExporter.log", append, %input );
		}
	} // End of loop

	echo( "Terrain successfully exported to text file." );

	// Cool sound effects
	playSound( 0, "bptSlct.wav", IDPRF_2D);
	schedule( "playSound( 0, \"dlink_complete.wav\", IDPRF_2D);" , 0.5);

	fileWrite( "TerrainExporter.log", append, strCat( "***END TERRAIN HEIGHTS*** ", $TerrainExporter::LogInfo, getTime()) );
}

// Taken from TrigonometryStdLib.cs
function raiseToPower(%number, %exponent)
{
   if(isInteger(%exponent) == false)
   {
      echo("raiseToPower() - Exponents must be integers!");
      %exponent = round(%exponent);
      echo("Exponent rounded to " @ %exponent);
   }
   %answer = %number;
   if(%exponent == 0)
   {
      %answer = 1;
   }
   else if(%exponent == 1)
   {
      %answer = %number;
   }
   else if(%exponent < 0)
   {
      if(%exponent == -1)
      {
         %answer = (1 / %number);
      }
      else if(%exponent <= -2)
      {
         %count = -2;
         while(%count >= %exponent)
         {
            %count--;
            %answer = (%answer * %number); 
         }
         %answer = (1 / %answer);
      }
   }
   else if(%exponent >= 2)
   {
      %count = 2;
      while(%count <= %exponent)
      {
         %count++;
         %answer = (%answer * %number); 
      }
   }
   return %answer;
}

function isInteger(%number)
{
   %excess = (%number - floor(%number));
   if(%excess == 0)
   {
      return true;
   }
   else
   {
      return false;
   }
}

function round(%number)
{
   %roundNum = %number;
   if(%number > 0)
   {
      %excess = (%number - floor(%number));
      if(%excess >= 0.5)
      {
         %roundNum = (floor(%number) + 1);
      }
      else
      {
         %roundNum = floor(%number);
      }
   }
   if(%number < 0)
   {
      %excess = (floor(%number) - %number);
      if(%excess <= 0.5)
      {
         %roundNum = floor(%number);
      }
      else
      {
         %roundNum = (floor(%number) - 1);
      }
   }
   return %roundNum;
}